#include "NativePCMDevice.h"
#include "error_codes.h"

#include "ExampleSourceLine.h"
#include "ExampleTargetLine.h"
#include "Line.h"

#include "stdio.h"
#include "string.h"

static const char* DEVICE_NAME = "Native Example Device";
static const char* DEVICE_VERSION = "0.1";

static const uint32 NUM_LINES = 4;

static Line* LINE[NUM_LINES] = {
	new ExampleTargetLine("Output Line A"), 
	new ExampleTargetLine("Output Line B"), 
	new ExampleSourceLine("Input Line A"), 
	new ExampleSourceLine("Input Line B")};

static const uint32 LINE_UIDS[NUM_LINES] = {
		130,
		1314,
		441,
		51};

DEV_ERR_CODE GetLineForUID(uint32 lineUID, Line **line)
{
	for (int i=0; i<NUM_LINES; i++)
	{
		if (LINE_UIDS[i] == lineUID)
		{
			*line = LINE[i];
			return NO_ERROR;
		}
	}
	return ILLEGAL_LINE_UID;
}

/**
 * Opens the line with lineUID of the output device.
 */
DEV_ERR_CODE Native_Line_Open(uint32 lineUID)
{
	Line *line = NULL;
	DEV_ERR_CODE err = GetLineForUID(lineUID, &line);
	if (err != NO_ERROR)
	{
		return err;
	}
	return line->Open();
}

/**
 * Query if the device is opened
 */
DEV_ERR_CODE Native_Line_IsOpen(uint32 lineUID, bool8 *isOpen)
{
	Line *line = NULL;
	DEV_ERR_CODE err = GetLineForUID(lineUID, &line);
	if (err != NO_ERROR)
	{
		return err;
	}
	*isOpen = line->IsOpen();
	return NO_ERROR;
}


/**
 * Query if the device is started
 */
DEV_ERR_CODE Native_Line_IsStarted(uint32 lineUID, bool8 *isStarted)
{
	Line *line = NULL;
	DEV_ERR_CODE err = GetLineForUID(lineUID, &line);
	if (err != NO_ERROR)
	{
		return err;
	}
	*isStarted = line->IsStarted();
	return NO_ERROR;
}


/**
 * Closes the line with lineUID of the output device. 
 */
DEV_ERR_CODE Native_Line_Close(uint32 lineUID)
{
	Line *line = NULL;
	DEV_ERR_CODE err = GetLineForUID(lineUID, &line);
	if (err != NO_ERROR)
	{
		return err;
	}
	return line->Close();
}


/**
 * Starts the line with lineUID of the output device.
 */
DEV_ERR_CODE Native_Line_Start(uint32 lineUID)
{
	Line *line = NULL;
	DEV_ERR_CODE err = GetLineForUID(lineUID, &line);
	if (err != NO_ERROR)
	{
		return err;
	}
	return line->Start();
}


/**
 * Stops the line with lineUID of the output device. An output device once stopped
 * might be started again.
 */
DEV_ERR_CODE Native_Line_Stop(uint32 lineUID)
{
	Line *line = NULL;
	DEV_ERR_CODE err = GetLineForUID(lineUID, &line);
	if (err != NO_ERROR)
	{
		return err;
	}
	return line->Stop();
}


/**
 * This function can be used to get the number of 
 * writable bytes in the output buffer of
 * the line with lineUID.
 * If the output device buffer is full, numWritable is equal to zero.
 */
DEV_ERR_CODE Native_Line_GetNumBytesReadWrite(uint32 lineUID, uint32 *numBytesWriteRead)
{
	Line *line = NULL;
	DEV_ERR_CODE err = GetLineForUID(lineUID, &line);
	if (err != NO_ERROR)
	{
		return err;
	}
	return line->GetNumReadWrite(numBytesWriteRead);
}


/**
 * Write the samples at *pBytes to the output device line
 * with lineUID. The number of bytes written to the
 * output device line is returned by numWritten.
 * The sample format is as follows:
 * 8 bit  (1 byte per sample)
 * 16 bit (2 bytes per sample)
 * 24 bit (3 bytes per sample)
 * 32 bit (4 bytes per sample)
 * If there is not enough buffer space available to write to the 
 * output buffer, the write method is allowed to block.
 */
DEV_ERR_CODE Native_Line_ReadWrite(uint32 lineUID, char pBytes[], uint32 numBytes, bool8 isSigned, uint32 *numReadWritten)
{
	Line *line = NULL;
	DEV_ERR_CODE err = GetLineForUID(lineUID, &line);
	if (err != NO_ERROR)
	{
		return err;
	}
	return line->ReadWrite(pBytes, numBytes, isSigned, numReadWritten);
}


/**
 * This function returns the length of 
 * the specific line output buffer in bytes.
 * NOTE: Buffer lengths < 500 ms are not recommended.
 */
DEV_ERR_CODE Native_Line_GetNumBytesBuffer(uint32 lineUID, uint32 *numBytes)
{
	Line *line = NULL;
	DEV_ERR_CODE err = GetLineForUID(lineUID, &line);
	if (err != NO_ERROR)
	{
		return err;
	}
	return line->GetNumBytesBuffer(numBytes);
}


/**
 * The name of the device which is shown in the output dialog of Spark is written to name.
 * The numChars parameter indicates the number of bytes available to write to name[].
 * NOTE: This function always returns the number of characters 
 *       required to fill the msg array.
 */
DEV_ERR_CODE Native_Device_GetName(char name[], uint32 numChars, uint32 *numCharsRequired)
{
	*numCharsRequired = (uint32)strlen(DEVICE_NAME)+1; // +1 Add zero delimiter '\0'
	if (numChars > *numCharsRequired)
	{
		numChars = *numCharsRequired;
	}
	strncpy(name, DEVICE_NAME, numChars);
	return NO_ERROR;
}


/**
 * The version of the device driver is written to the version array. 
 * The numChars parameter indicates the number of bytes
 * available to write to version[].
 * NOTE: This function always returns the number of characters 
 *       required to fill the version array incl. zero delimiter.
 */
DEV_ERR_CODE Native_Device_GetVersion(char version[], uint32 numChars, uint32 *numCharsRequired)
{
	*numCharsRequired = (uint32)strlen(DEVICE_VERSION)+1; // +1 Add zero delimiter '\0'
	if (numChars > *numCharsRequired)
	{
		numChars = *numCharsRequired;
	}
	strncpy(version, DEVICE_VERSION, numChars);
	return NO_ERROR;
}


/**
 * The name of the line with lineUID is written to the name array.
 * The numChars parameter indicates the number of bytes
 * available to write to name[].
 * NOTE: This function always returns the number of characters 
 *       required to fill the name array incl. zero delimiter.
 */
DEV_ERR_CODE Native_Line_GetName(uint32 lineUID, char name[], uint32 numChars, uint32 *numCharsRequired)
{
	Line *line = NULL;
	DEV_ERR_CODE err = GetLineForUID(lineUID, &line);
	if (err != NO_ERROR)
	{
		return err;
	}
	line->GetName(name, numChars, numCharsRequired);
	return NO_ERROR;
}


/**
 * Returns the UIDs of all lines available on this device.
 */
DEV_ERR_CODE Native_Device_GetLineUIDs(uint32 uids[], uint32 len)
{
	if (len > NUM_LINES) len = NUM_LINES;
	for (uint32 i=0; i<len; i++)
	{
		uids[i] = LINE_UIDS[i];
	}
	return NO_ERROR;
}


/**
 * Returns the number of lines available
 */
DEV_ERR_CODE Native_Device_GetNumLines(uint32 *numLines)
{
	*numLines = NUM_LINES;
	return NO_ERROR;
}


/**
 * The user can query the error message corresponding to a certain error code
 * by calling this function. 
 * The error message corresponing the the error code is written to the msg array.
 * The numChars parameter indicates the number of bytes
 * available to write to msg[].
 * NOTE: This function always returns the number of characters 
 *       required to fill the msg array.
 */
void Native_GetErrorMsg(DEV_ERR_CODE code, char msg[], uint32 numChars, uint32 *numCharsRequired)
{

}


/**
 * Returns true (1) if the error code in the argument indicates an error.
 * Returns false (0) otherwise.
 */
bool8 Native_IsError(DEV_ERR_CODE code)
{
	return code > NO_ERROR;
}


/**
 * Returns true (1) if the error code in the argument indicates a warning.
 * Returns false (0) otherwise.
 */
bool8 Native_IsWarning(DEV_ERR_CODE code)
{
	return code < NO_ERROR;
}


/**
 * Returns the samplerate of the output line with lineUID in Hz.
 * It is recommended to use samplerates in multiples of 1 kHz (1000 Hz).
 * (e.g. 93000 Hz)
 */
DEV_ERR_CODE Native_Line_GetSampleRateHz(uint32 lineUID, uint32 *sampleRateHz)
{
	Line *line = NULL;
	DEV_ERR_CODE err = GetLineForUID(lineUID, &line);
	if (err != NO_ERROR)
	{
		return err;
	}
	*sampleRateHz = line->GetSampleRateHz();
	return NO_ERROR;
}


/**
 * Returns the number of bytes per sample of the output line with lineUID. (e.g. 16 bit resolution -> 2 bytes per sample)
 * It is recommended to use values between 1 and 4.
 */
DEV_ERR_CODE Native_Line_GetNumBytesPerSample(uint32 lineUID, uint32 *numBytesPerSample)
{
	Line *line = NULL;
	DEV_ERR_CODE err = GetLineForUID(lineUID, &line);
	if (err != NO_ERROR)
	{
		return err;
	}
	*numBytesPerSample = line->GetNumBytesPerSample();
	return NO_ERROR;
}


/**
 * Returns the number of channels of the output line with lineUID.
 * NOTE:
 * In case of I/Q output, the number of channels is 2 (complex signal).
 * In case of I output only, the number of channels is 1 (real signal).
 * Values other than 1 or 2 are not permitted.
 */
DEV_ERR_CODE Native_Line_GetNumChannels(uint32 lineUID, uint32 *numChannels)
{
	Line *line = NULL;
	DEV_ERR_CODE err = GetLineForUID(lineUID, &line);
	if (err != NO_ERROR)
	{
		return err;
	}
	*numChannels = line->GetNumChannels();
	return NO_ERROR;
}


/**
 * Writes the line type to *lineType. If the line is a source line, i.e. the samples
 * are read from the device, the line type is SOURCE_LINE.
 * If the line is a target line, i.e. the samples are written to the device
 * the line type is a TARGET_LINE.
 */
DEV_ERR_CODE Native_Line_GetType(uint32 lineUID, LineType *lineType)
{
	Line *line = NULL;
	DEV_ERR_CODE err = GetLineForUID(lineUID, &line);
	if (err != NO_ERROR)
	{
		return err;
	}
	*lineType = line->GetType();
	return NO_ERROR;
}
